from graphlib import TopologicalSorter
from pathlib import Path
from typing import Sequence

from jinja2 import Environment, FileSystemLoader

from spec2sdk.generators.converters import converters
from spec2sdk.generators.entities import PythonType
from spec2sdk.parsers.entities import Specification

from ..imports import render_imports
from ..utils import get_root_data_types
from .entities import TypeRenderer
from .renderers import renderers


def create_jinja_environment() -> Environment:
    return Environment(loader=FileSystemLoader(f"{Path(__file__).parent}/templates"))


def unwrap_py_types(py_types: Sequence[PythonType]) -> Sequence[PythonType]:
    """
    Recursively unwraps all python types and returns all unique python types.
    """
    return tuple(
        {
            dependent_type
            for py_type in py_types
            for dependent_type in (py_type, *unwrap_py_types(py_type.dependency_types))
        },
    )


def generate_models(spec: Specification) -> str:
    root_data_types = get_root_data_types(spec)
    root_py_types = tuple(map(converters.convert, root_data_types))
    all_py_types = unwrap_py_types(root_py_types)

    # Types must be sorted in the order of defining their dependencies
    topological_sorter = TopologicalSorter({py_type: py_type.dependency_types for py_type in all_py_types})
    topological_sorter.prepare()
    type_renderers: Sequence[TypeRenderer] = ()

    while topological_sorter.is_active():
        node_group = topological_sorter.get_ready()
        type_renderers += tuple(map(renderers.convert, sorted(node_group, key=lambda py_type: py_type.name or "")))
        topological_sorter.done(*node_group)

    # Render imports
    content = render_imports(
        tuple({import_ for type_renderer in type_renderers for import_ in type_renderer.imports}),
    )

    # Render base model class
    content += create_jinja_environment().get_template("base_model.j2").render() + "\n"

    # Render types with a name
    content += "\n".join(type_renderer.content for type_renderer in type_renderers if type_renderer.content is not None)

    return content



from typing import Any, Callable, Sequence, TypedDict

from ..registry import Registry
from .entities import (
    AllOfDataType,
    AnyOfDataType,
    ArrayDataType,
    BooleanDataType,
    Content,
    Endpoint,
    Enumerator,
    IntegerDataType,
    NumberDataType,
    ObjectDataType,
    ObjectProperty,
    OneOfDataType,
    Parameter,
    Path,
    RequestBody,
    Response,
    Specification,
    StringDataType,
)
from .exceptions import ParserError
from .predicates import contains, type_equals
from .resolver import SCHEMA_NAME_FIELD

parsers = Registry()


def parse_enumerators[V](schema: dict, type_parser: Callable[[V], V] | None = None) -> Sequence[Enumerator] | None:
    if enum_member_values := schema.get("enum"):
        enum_member_names = schema.get("x-enum-varnames", (None,) * len(enum_member_values))

        if len(enum_member_names) != len(enum_member_values):
            raise ValueError(
                f"{enum_member_values}: enum values count do not match the number of `x-enum-varnames` names",
            )

        return tuple(
            Enumerator(name=name, value=value if (value is None) or (type_parser is None) else type_parser(value))
            for name, value in zip(enum_member_names, enum_member_values)
        )
    else:
        return None


def parse_default_value[I, O](schema: dict, type_parser: Callable[[I], O] | None = None) -> Any | None:
    if "default" in schema:
        value = schema["default"]
        if (value is not None) and (type_parser is not None):
            value = type_parser(value)

        return value
    else:
        return None


class CommonFields(TypedDict):
    name: str | None
    description: str | None
    default_value: Any | None
    enumerators: Sequence[Enumerator] | None
    is_nullable: bool


def parse_common_fields[I, O](
    schema: dict,
    type_parser: Callable[[I], O] | None = None,
) -> CommonFields:
    return CommonFields(
        name=schema.get(SCHEMA_NAME_FIELD),
        description=schema.get("description"),
        default_value=parse_default_value(schema, type_parser),
        enumerators=parse_enumerators(schema, type_parser),
        is_nullable=schema.get("nullable") == "true",
    )


@parsers.register(predicate=type_equals("string"))
def parse_string(schema: dict) -> StringDataType:
    return StringDataType(
        **parse_common_fields(schema=schema, type_parser=str),
        format=schema.get("format"),
        pattern=schema.get("pattern"),
    )


@parsers.register(predicate=type_equals("number"))
def parse_number(schema: dict) -> NumberDataType:
    return NumberDataType(
        **parse_common_fields(schema=schema, type_parser=float),
        format=schema.get("format"),
    )


@parsers.register(predicate=type_equals("integer"))
def parse_integer(schema: dict) -> IntegerDataType:
    return IntegerDataType(
        **parse_common_fields(schema=schema, type_parser=int),
        format=schema.get("format"),
    )


@parsers.register(predicate=type_equals("boolean"))
def parse_boolean(schema: dict) -> BooleanDataType:
    return BooleanDataType(
        **parse_common_fields(schema=schema, type_parser=bool),
    )


@parsers.register(predicate=type_equals("object"))
def parse_object(schema: dict) -> ObjectDataType:
    additional_properties: bool = ("properties" not in schema) and schema.get("additionalProperties") in (
        None,
        True,
        {},
    )

    return ObjectDataType(
        **parse_common_fields(schema=schema),
        properties=tuple(
            ObjectProperty(
                data_type=parsers.convert(property_schema),
                name=property_name,
                is_required=property_name in schema.get("required", ()),
            )
            for property_name, property_schema in schema.get("properties", {}).items()
        ),
        additional_properties=additional_properties,
    )


@parsers.register(predicate=type_equals("array"))
def parse_array(schema: dict) -> ArrayDataType:
    return ArrayDataType(
        **parse_common_fields(schema=schema),
        item_type=parsers.convert(schema["items"]),
    )


@parsers.register(predicate=contains("oneOf"))
def parse_one_of(schema: dict) -> OneOfDataType:
    return OneOfDataType(
        **parse_common_fields(schema=schema),
        data_types=tuple(map(parsers.convert, schema["oneOf"])),
    )


@parsers.register(predicate=contains("anyOf"))
def parse_any_of(schema: dict) -> AnyOfDataType:
    return AnyOfDataType(
        **parse_common_fields(schema=schema),
        data_types=tuple(map(parsers.convert, schema["anyOf"])),
    )


@parsers.register(predicate=contains("allOf"))
def parse_all_of(schema: dict) -> AllOfDataType:
    return AllOfDataType(
        **parse_common_fields(schema=schema),
        data_types=tuple(map(parsers.convert, schema["allOf"])),
    )


def parse_spec(schema: dict) -> Specification:
    endpoints = []

    for path, operations in schema.get("paths", {}).items():
        for method, operation in operations.items():
            parameters = tuple(
                Parameter(
                    name=parameter["name"],
                    location=parameter["in"],
                    description=parameter.get("description"),
                    required=parameter.get("required", False),
                    data_type=parsers.convert(parameter["schema"]),
                    default_value=parameter["schema"].get("default"),
                )
                for parameter in operation.get("parameters", ())
            )

            if body := operation.get("requestBody"):
                if len(body["content"]) != 1:
                    raise ParserError(
                        f"Multiple content items in the request body are not supported: {method=} {path=}",
                    )

                media_type, content = next(iter(body["content"].items()))
                request_body = RequestBody(
                    description=body.get("description"),
                    required=body.get("required", False),
                    content=Content(
                        media_type=media_type,
                        data_type=parsers.convert(content["schema"]),
                    ),
                )
            else:
                request_body = None

            responses = []
            for status_code, response in operation["responses"].items():
                if "content" in response:
                    if len(response["content"]) != 1:
                        raise ParserError(
                            f"Multiple content items in the responses are not supported: "
                            f"{method=} {path=} {status_code=}",
                        )

                    media_type, content = next(iter(response["content"].items()))
                    responses.append(
                        Response(
                            status_code=status_code,
                            description=response["description"],
                            content=Content(
                                media_type=media_type,
                                data_type=parsers.convert(content["schema"]),
                            ),
                        ),
                    )
                else:
                    responses.append(
                        Response(
                            status_code=status_code,
                            description=response["description"],
                            content=None,
                        ),
                    )

            endpoints.append(
                Endpoint(
                    path=Path(path=path, parameters=parameters),
                    method=method,
                    operation_id=operation.get("operationId"),
                    summary=operation.get("summary"),
                    request_body=request_body,
                    responses=tuple(responses),
                ),
            )

    return Specification(
        endpoints=endpoints,
    )



from pathlib import Path
from typing import Sequence

from jinja2 import Environment, FileSystemLoader

from spec2sdk.generators.converters import converters
from spec2sdk.generators.entities import PythonType
from spec2sdk.parsers.entities import Specification

from ..imports import Import, render_imports
from ..utils import get_root_data_types
from .views import EndpointView


def get_imports(py_types: Sequence[PythonType], models_import: Import) -> Sequence[Import]:
    def get_importable_type_names(py_type: PythonType) -> Sequence[str]:
        return (
            (py_type.name,)
            if py_type.name
            else tuple(
                type_name
                for dependency_type in py_type.dependency_types
                for type_name in get_importable_type_names(dependency_type)
            )
        )

    return tuple(
        {
            models_import.model_copy(update={"name": name})
            for py_type in py_types
            for name in get_importable_type_names(py_type)
        },
    )


def generate_client(spec: Specification, models_import: Import) -> str:
    root_data_types = get_root_data_types(spec)
    root_python_types = tuple(map(converters.convert, root_data_types))
    imports = get_imports(py_types=root_python_types, models_import=models_import)

    return (
        Environment(loader=FileSystemLoader(f"{Path(__file__).parent}/templates"))
        .get_template("client.j2")
        .render(
            imports=render_imports(imports),
            endpoints=tuple(
                EndpointView(endpoint=endpoint, response=response)
                for endpoint in spec.endpoints
                for response in endpoint.responses
            ),
        )
    )



from typing import Callable


def contains(key: str) -> Callable[[dict], bool]:
    def compare(schema: dict) -> bool:
        return key in schema

    return compare


def type_equals(type_name: str) -> Callable[[dict], bool]:
    def compare(schema: dict) -> bool:
        return schema.get("type") == type_name

    return compare



from pathlib import Path
from typing import Sequence

from jinja2 import Environment, FileSystemLoader

from spec2sdk.base import Model


class Import(Model):
    name: str
    package: str


def render_imports(imports: Sequence[Import]) -> str:
    jinja_environment = Environment(loader=FileSystemLoader(f"{Path(__file__).parent}/templates"))
    return jinja_environment.get_template("imports.j2").render(imports=imports)



from typing import Sequence

from spec2sdk.parsers.entities import DataType, Specification


def get_root_data_types(spec: Specification) -> Sequence[DataType]:
    """
    Returns the only data types used directly in the parameters, request bodies and responses.
    Duplicates are removed from the result.
    """
    data_types = set()

    for endpoint in spec.endpoints:
        data_types.update(tuple(parameter.data_type for parameter in endpoint.path.parameters))

        if endpoint.request_body:
            data_types.add(endpoint.request_body.content.data_type)

        data_types.update(tuple(response.content.data_type for response in endpoint.responses if response.content))

    return tuple(data_types)



from typing import Any, Self, Sequence

from spec2sdk.base import Model


class PythonType(Model):
    name: str | None
    type_hint: str | None
    description: str | None
    default_value: Any

    @property
    def dependency_types(self) -> Sequence[Self]:
        return ()

    def __hash__(self):
        return hash((self.__class__, self.name or self.type_hint))

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, self.__class__):
            return self.name == other.name
        else:
            return self.model_dump() == other


class LiteralPythonType(PythonType):
    literals: Sequence[Any]


class EnumMember(Model):
    name: str
    value: Any


class EnumPythonType(PythonType):
    name: str
    members: Sequence[EnumMember]
    default_value: EnumMember | None


class StrEnumPythonType(EnumPythonType):
    pass


class IntegerPythonType(PythonType):
    default_value: int | None


class FloatPythonType(PythonType):
    default_value: float | None


class BooleanPythonType(PythonType):
    default_value: bool | None


class StringPythonType(PythonType):
    default_value: str | None


class FilePythonType(PythonType):
    pass


class ModelField(Model):
    name: str
    alias: str
    type_hint: str | None
    description: str | None
    default_value: Any
    is_required: bool
    inner_py_type: PythonType


class ModelPythonType(PythonType):
    base_models: Sequence["ModelPythonType"]
    fields: Sequence[ModelField]
    arbitrary_fields_allowed: bool

    @property
    def dependency_types(self) -> Sequence[PythonType]:
        return *tuple(field.inner_py_type for field in self.fields), *self.base_models


class OptionalPythonType(PythonType):
    inner_py_type: PythonType

    @property
    def dependency_types(self) -> Sequence[PythonType]:
        return (self.inner_py_type,)


class ListPythonType(PythonType):
    inner_py_type: PythonType

    @property
    def dependency_types(self) -> Sequence[PythonType]:
        return (self.inner_py_type,)


class UnionPythonType(PythonType):
    inner_py_types: Sequence[PythonType]

    @property
    def dependency_types(self) -> Sequence[PythonType]:
        return self.inner_py_types



from typing import Any, TypedDict

from spec2sdk.generators.entities import (
    BooleanPythonType,
    EnumMember,
    EnumPythonType,
    FilePythonType,
    FloatPythonType,
    IntegerPythonType,
    ListPythonType,
    LiteralPythonType,
    ModelField,
    ModelPythonType,
    OptionalPythonType,
    StrEnumPythonType,
    StringPythonType,
    UnionPythonType,
)
from spec2sdk.generators.identifiers import make_class_name, make_constant_name, make_variable_name
from spec2sdk.generators.predicates import is_binary_format, is_enum, is_instance, is_literal, is_optional, is_str_enum
from spec2sdk.parsers.entities import (
    AllOfDataType,
    AnyOfDataType,
    ArrayDataType,
    BooleanDataType,
    DataType,
    Enumerator,
    IntegerDataType,
    NumberDataType,
    ObjectDataType,
    OneOfDataType,
    StringDataType,
)
from spec2sdk.registry import Registry

converters = Registry()


class CommonFields(TypedDict):
    name: str | None
    type_hint: str | None
    description: str | None
    default_value: Any


def get_common_fields(data_type: DataType, type_hint: str | None = None) -> CommonFields:
    return CommonFields(
        name=make_class_name(data_type.name) if data_type.name else None,
        type_hint=make_class_name(data_type.name) if data_type.name else type_hint,
        description=data_type.description,
        default_value=data_type.default_value,
    )


@converters.register(predicate=is_instance(StringDataType))
def convert_string(data_type: StringDataType) -> StringPythonType:
    return StringPythonType(
        **get_common_fields(data_type, "str"),
    )


@converters.register(predicate=is_instance(IntegerDataType))
def convert_integer(data_type: IntegerDataType) -> IntegerPythonType:
    return IntegerPythonType(
        **get_common_fields(data_type, "int"),
    )


@converters.register(predicate=is_instance(NumberDataType))
def convert_number(data_type: NumberDataType) -> FloatPythonType:
    return FloatPythonType(
        **get_common_fields(data_type, "float"),
    )


@converters.register(predicate=is_instance(BooleanDataType))
def convert_boolean(data_type: BooleanDataType) -> BooleanPythonType:
    return BooleanPythonType(
        **get_common_fields(data_type, "bool"),
    )


@converters.register(predicate=is_instance(ObjectDataType))
def convert_object(data_type: ObjectDataType) -> ModelPythonType:
    fields = []

    for prop in data_type.properties:
        property_data_type = prop.data_type

        if not prop.is_required:
            property_data_type = property_data_type.model_copy(update={"is_nullable": True})

        inner_py_type = converters.convert(property_data_type)

        fields.append(
            ModelField(
                name=make_variable_name(prop.name),
                alias=prop.name,
                type_hint=inner_py_type.type_hint,
                description=prop.data_type.description,
                default_value=inner_py_type.default_value,
                is_required=prop.is_required,
                inner_py_type=inner_py_type,
            ),
        )

    return ModelPythonType(
        **get_common_fields(data_type),
        base_models=(),
        fields=tuple(fields),
        arbitrary_fields_allowed=data_type.additional_properties,
    )


@converters.register(predicate=is_instance(ArrayDataType))
def convert_array(data_type: ArrayDataType) -> ListPythonType:
    inner_py_type = converters.convert(data_type.item_type)

    return ListPythonType(
        **get_common_fields(data_type, f"list[{inner_py_type.type_hint}]"),
        inner_py_type=inner_py_type,
    )


@converters.register(predicate=is_instance((AnyOfDataType, OneOfDataType)))
def convert_one_of(data_type: OneOfDataType | AnyOfDataType) -> UnionPythonType:
    inner_py_types = tuple(map(converters.convert, data_type.data_types))

    return UnionPythonType(
        **get_common_fields(data_type, " | ".join(py_type.type_hint for py_type in inner_py_types)),
        inner_py_types=inner_py_types,
    )


@converters.register(predicate=is_instance(AllOfDataType))
def convert_all_of(data_type: AllOfDataType) -> ModelPythonType:
    if not all(isinstance(inner_data_type, ObjectDataType) for inner_data_type in data_type.data_types):
        raise TypeError("Non-object data types in allOf is not supported")

    nameless_inner_data_types = tuple(
        inner_data_type for inner_data_type in data_type.data_types if inner_data_type.name is None
    )
    if len(nameless_inner_data_types) > 1:
        raise TypeError("Multiple data types without a name in allOf is not supported")

    model_type = convert_object(
        nameless_inner_data_types[0] if nameless_inner_data_types else data_type.data_types[0],
    )

    return ModelPythonType(
        name=make_class_name(data_type.name),
        type_hint=make_class_name(data_type.name),
        description=data_type.description,
        default_value=model_type.default_value,
        base_models=tuple(
            converters.convert(inner_data_type)
            for inner_data_type in data_type.data_types
            if inner_data_type.name is not None
        ),
        fields=model_type.fields if nameless_inner_data_types else (),
        arbitrary_fields_allowed=False,
    )


@converters.register(predicate=is_enum)
def convert_enum(data_type: DataType) -> EnumPythonType:
    def generate_enum_member_name(enumerator: Enumerator) -> str:
        if enumerator.name:
            name = enumerator.name
        elif isinstance(enumerator.value, str):
            name = enumerator.value
        else:
            name = f"{data_type.name}_{enumerator.value}"

        return make_constant_name(name)

    members = tuple(
        EnumMember(name=generate_enum_member_name(member), value=member.value) for member in data_type.enumerators
    )

    default_value = None
    for member in members:
        if member.value == data_type.default_value:
            default_value = member

    return EnumPythonType(
        name=make_class_name(data_type.name),
        type_hint=make_class_name(data_type.name),
        description=data_type.description,
        default_value=default_value,
        members=members,
    )


@converters.register(predicate=is_str_enum)
def convert_str_enum(data_type: StringDataType) -> StrEnumPythonType:
    enum_type = convert_enum(data_type)
    return StrEnumPythonType(
        name=enum_type.name,
        type_hint=enum_type.type_hint,
        description=enum_type.description,
        default_value=enum_type.default_value,
        members=enum_type.members,
    )


@converters.register(predicate=is_binary_format)
def convert_binary(data_type: StringDataType) -> FilePythonType:
    return FilePythonType(
        name="File",
        type_hint="File",
        description=data_type.description,
        default_value=data_type.default_value,
    )


@converters.register(predicate=is_literal)
def convert_literal(data_type: DataType) -> LiteralPythonType:
    literals = tuple(enumerator.value for enumerator in data_type.enumerators)
    return LiteralPythonType(
        name=None,
        type_hint="Literal[" + ",".join(repr(literal) for literal in literals) + "]",
        description=data_type.description,
        default_value=data_type.default_value,
        literals=literals,
    )


@converters.register(predicate=is_optional, priority=99)
def convert_optional(data_type: DataType) -> OptionalPythonType:
    inner_py_type = converters.convert(data_type.model_copy(update={"is_nullable": False}))

    return OptionalPythonType(
        name=None,
        type_hint=f"{inner_py_type.type_hint} | None",
        description=data_type.description,
        default_value=data_type.default_value,
        inner_py_type=inner_py_type,
    )



import re
from keyword import iskeyword

from humps import decamelize, pascalize


def make_identifier(name: str) -> str:
    """
    Makes valid Python identifier from the string.
    """
    # Remove invalid characters
    name = re.sub(r"[^0-9a-zA-Z_]", "_", name)

    # Remove leading characters until we find a letter
    name = re.sub(r"^[^a-zA-Z]+", "", name)

    # Replace consecutive duplicates of the underscore
    name = re.sub(r"_+", "_", name)

    # Add underscore to the name if it's a valid Python keyword
    if iskeyword(name):
        name += "_"

    return name


def make_class_name(name: str) -> str:
    return pascalize(make_identifier(name))


def make_constant_name(name: str) -> str:
    return make_identifier(name).upper()


def make_variable_name(name: str) -> str:
    return decamelize(make_identifier((name)))



import argparse
import shutil
import subprocess
from pathlib import Path
from urllib.parse import urlparse

from openapi_spec_validator import validate

from spec2sdk.generators.client.generators import generate_client
from spec2sdk.generators.imports import Import
from spec2sdk.generators.models.generators import generate_models
from spec2sdk.parsers.parsers import parse_spec
from spec2sdk.parsers.resolver import ResolvingParser


def format_file(path: Path):
    def run_formatter(formatter: str):
        print(f"Running {formatter} on {path}")
        subprocess.run(f"{formatter} {path}", shell=True, check=True)

    # Formatting and linting explanation:
    # 1. Format the code [line will be a 120 character]
    # 2. Add trailing comma [line will be a 121 character]
    # 3. Format the code [line will be a 120 character]
    # 4. Apply linting auto fixes, and then check if the generated code follow the linting rules
    run_formatter("ruff format")
    run_formatter("ruff check --select COM812 --fix")
    run_formatter("ruff format")
    run_formatter("ruff check --fix")


def generate(url: str, output_dir: Path):
    schema = ResolvingParser().parse(url=url)
    validate(schema)
    spec = parse_spec(schema)

    if output_dir.exists():
        shutil.rmtree(str(output_dir))

    output_dir.mkdir(parents=True, exist_ok=True)
    output_dir.joinpath("__init__.py").write_text("")

    models_path = output_dir.joinpath("models.py")
    models_path.write_text(generate_models(spec=spec))
    format_file(models_path)

    client_path = output_dir.joinpath("api_client.py")
    client_path.write_text(generate_client(spec=spec, models_import=Import(name="", package=f".{models_path.stem}")))
    format_file(client_path)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--input",
        type=str,
        help="File path or URL to the OpenAPI specification file in the YAML format",
        required=True,
    )
    parser.add_argument(
        "--output-dir",
        type=Path,
        help="Path to the output directory where the generated code will be written to",
        required=True,
    )

    args = parser.parse_args()
    generate(
        url=Path(args.input).absolute().as_uri() if urlparse(args.input).scheme == "" else args.input,
        output_dir=args.output_dir,
    )


if __name__ == "__main__":
    main()



import textwrap
from pathlib import Path
from typing import Sequence

from jinja2 import Environment, FileSystemLoader

from spec2sdk.generators.entities import (
    BooleanPythonType,
    EnumPythonType,
    FilePythonType,
    FloatPythonType,
    IntegerPythonType,
    ListPythonType,
    LiteralPythonType,
    ModelField,
    ModelPythonType,
    OptionalPythonType,
    PythonType,
    StrEnumPythonType,
    StringPythonType,
    UnionPythonType,
)
from spec2sdk.generators.models.entities import TypeRenderer
from spec2sdk.generators.predicates import is_instance
from spec2sdk.registry import Registry

from ...base import Model
from ..imports import Import

renderers = Registry()


def create_jinja_environment() -> Environment:
    return Environment(loader=FileSystemLoader(f"{Path(__file__).parent}/templates"))


def render_root_model(
    py_type: PythonType,
    extra_imports: Sequence[Import] = (),
    content: str = "",
) -> TypeRenderer:
    return TypeRenderer(
        imports=(
            *extra_imports,
            *((Import(name="RootModel", package="pydantic"),) if py_type.name else ()),
        ),
        content=f"{py_type.name} = RootModel[{content}]" if py_type.name else None,
    )


@renderers.register(predicate=is_instance(IntegerPythonType))
def render_integer(py_type: IntegerPythonType) -> TypeRenderer:
    return render_root_model(py_type, content="int")


@renderers.register(predicate=is_instance(FloatPythonType))
def render_float(py_type: FloatPythonType) -> TypeRenderer:
    return render_root_model(py_type, content="float")


@renderers.register(predicate=is_instance(BooleanPythonType))
def render_boolean(py_type: BooleanPythonType) -> TypeRenderer:
    return render_root_model(py_type, content="bool")


@renderers.register(predicate=is_instance(StringPythonType))
def render_string(py_type: StringPythonType) -> TypeRenderer:
    return render_root_model(py_type, content="str")


@renderers.register(predicate=is_instance(OptionalPythonType))
def render_optional(py_type: OptionalPythonType) -> TypeRenderer:
    return render_root_model(
        py_type=py_type,
        content=f"{py_type.inner_py_type.type_hint} | None",
    )


@renderers.register(predicate=is_instance(ListPythonType))
def render_list(py_type: ListPythonType) -> TypeRenderer:
    return render_root_model(
        py_type=py_type,
        content=f"list[{py_type.inner_py_type.type_hint}]",
    )


@renderers.register(predicate=is_instance(FilePythonType))
def render_file(py_type: FilePythonType) -> TypeRenderer:
    return TypeRenderer(
        imports=(Import(name="NamedTuple", package="typing"),),
        content="""
class File(NamedTuple):
    name: str
    content: bytes
"""
        if py_type.name is not None
        else None,
    )


@renderers.register(predicate=is_instance(LiteralPythonType))
def render_literal(py_type: LiteralPythonType) -> TypeRenderer:
    return render_root_model(
        py_type=py_type,
        extra_imports=(Import(name="Literal", package="typing"),),
        content="Literal[" + ",".join(repr(literal) for literal in py_type.literals) + "]",
    )


class EnumMemberView(Model):
    name: str
    value: str


@renderers.register(predicate=is_instance(EnumPythonType))
def render_enum(py_type: EnumPythonType) -> TypeRenderer:
    return TypeRenderer(
        imports=(Import(name="Enum", package="enum"),),
        content=(
            create_jinja_environment()
            .get_template("enum.j2")
            .render(
                enum_type=py_type,
                base_class_name="Enum",
                members=tuple(EnumMemberView(name=member.name, value=member.value) for member in py_type.members),
            )
            if py_type.name is not None
            else None
        ),
    )


@renderers.register(predicate=is_instance(StrEnumPythonType))
def render_str_enum(py_type: StrEnumPythonType) -> TypeRenderer:
    return TypeRenderer(
        imports=(Import(name="StrEnum", package="enum"),),
        content=(
            create_jinja_environment()
            .get_template("enum.j2")
            .render(
                enum_type=py_type,
                base_class_name="StrEnum",
                members=tuple(
                    EnumMemberView(name=member.name, value=f'"{member.value}"') for member in py_type.members
                ),
            )
            if py_type.name is not None
            else None
        ),
    )


@renderers.register(predicate=is_instance(UnionPythonType))
def render_union(py_type: UnionPythonType) -> TypeRenderer:
    return render_root_model(
        py_type,
        content=" | ".join(py_type.type_hint for py_type in py_type.inner_py_types),
    )


class ModelFieldView(Model):
    name: str
    type_definition: str


@renderers.register(predicate=is_instance(ModelPythonType))
def render_model(py_type: ModelPythonType) -> TypeRenderer:
    def split_long_lines(s: str) -> str:
        return '"' + '""'.join(line.replace('"', r"\"") for line in textwrap.wrap(s, width=80)) + '"'

    def create_model_field_view(field: ModelField) -> ModelFieldView:
        attrs = []

        if field.default_value is not None or not field.is_required:
            attrs.append(f"default={repr(field.default_value)}")

        if field.name != field.alias:
            attrs.append(f'alias="{field.alias}"')

        if field.description:
            attrs.append(f"description={split_long_lines(field.description)}")

        return ModelFieldView(
            name=field.name,
            type_definition=field.inner_py_type.type_hint + (f' = Field({",".join(attrs)})' if attrs else ""),
        )

    base_class_names = tuple(base_model.name for base_model in py_type.base_models if base_model.name)

    return TypeRenderer(
        imports=(
            *((Import(name="Field", package="pydantic"),) if len(py_type.fields) > 0 else ()),
            *((Import(name="ConfigDict", package="pydantic"),) if py_type.arbitrary_fields_allowed else ()),
        ),
        content=(
            create_jinja_environment()
            .get_template("model.j2")
            .render(
                base_class_name=", ".join(base_class_names) if base_class_names else "Model",
                model_type=py_type,
                fields=tuple(map(create_model_field_view, py_type.fields)),
                arbitrary_fields_allowed=py_type.arbitrary_fields_allowed,
            )
            if py_type.name
            else None
        ),
    )
